/*
 * Copyright 1996 John D. Polstra.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by John Polstra.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *      $Id: rtld.c,v 1.6 1996/03/15 01:39:12 jdp Exp $
 */

/*
 * Dynamic linker for ELF.
 *
 * John Polstra <jdp@polstra.com>.
 */

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "debug.h"
#include "rtld.h"

/*
 * Debugging support.
 */

#define str(x)	#x
#define xstr(x)	str(x)

#define assert(cond)	((cond) ? (void) 0 :\
    (msg("oops: " xstr(__LINE__) "\n"), abort()))	/* FIXME */
#define msg(s)		(write(1, s, strlen(s)))
#define trace()		msg("trace: " xstr(__LINE__) "\n");

/*
 * Function declarations.
 */
static const char *basename(const char *);
static void digest_dynamic(Obj_Entry *);
static Obj_Entry *digest_phdr(const Elf32_Phdr *, int, caddr_t);
static int do_copy_relocations(Obj_Entry *);
static unsigned long elf_hash(const char *);
static char *find_library(const char *, const Obj_Entry *);
static const Elf32_Sym *find_symdef(unsigned long, const Obj_Entry *,
    const Obj_Entry **, bool);
static void init_rtld(caddr_t);
static int load_needed_objects(Obj_Entry *);
static Obj_Entry *load_object(char *);
static int patch_addr(const char *, Elf32_Addr);
static int relocate_objects(Obj_Entry *);
static void rtld_exit(void);
static char *search_library_path(const char *, const char *);
static const Elf32_Sym *symlook(const char *, const Obj_Entry **, bool);
static const Elf32_Sym *symlook_obj(const char *, unsigned long,
    const Obj_Entry *, bool);

void xprintf(const char *, ...);

/* Assembly language entry point for lazy binding. */
extern void _rtld_bind_start(void);

/*
 * Assembly language macro for getting the GOT pointer.
 */
#define get_got_address(r)	__asm__( "movl %%ebx,%0" : "=rm"(r) )

/*
 * Data declarations.
 */
static bool trust;		/* False for setuid and setgid programs */
static char *ld_bind_now;	/* Environment variable for immediate binding */
static char *ld_debug;		/* Environment variable for debugging */
static char *ld_library_path;	/* Environment variable for search path */
static Obj_Entry *obj_list;	/* Head of linked list of shared objects */
static Obj_Entry **obj_tail;	/* Link field of last object in list */

/*
 * Global declarations normally provided by crt0.
 */
char *__progname;
char **environ;

/*
 * Main entry point for dynamic linking.  The argument is the stack
 * pointer.  The stack is expected to be laid out as described in the
 * SVR4 ABI specification, Intel 386 Processor Supplement.  Specifically,
 * the stack pointer points to a word containing ARGC.  Following that
 * in the stack is a null-terminated sequence of pointers to argument
 * strings.  Then comes a null-terminated sequence of pointers to
 * environment strings.  Finally, there is a sequence of "auxiliary
 * vector" entries.
 *
 * This function returns the entry point for the main program in %eax,
 * and the dynamic linker's exit procedure in %edx.  We accomplish this
 * by declaring the return value to have the 64-bit type "long long".
 * Such values are returned with their most-significant 32 bits in %edx,
 * and their least-significant 32 bits in %eax.
 */
long long
_rtld(Elf32_Word *sp)
{
    auxv_t *aux_info[AT_COUNT];
    int i;
    char **env;
    auxv_t *aux;
    auxv_t *auxp;
    Obj_Entry *obj_main;

    /*
     * On entry, the dynamic linker itself has not been relocated yet.
     * Be very careful not to reference any global data until after
     * init_rtld has returned.
     */

    /* Find the auxiliary vector on the stack. */
    sp += *sp + 2;	/* Skip over argc, arguments, and NULL terminator */
    env = (char **) sp;
    while(*sp++ != 0)	/* Skip over environment, and NULL terminator */
	;
    aux = (auxv_t *) sp;

    /* Digest the auxiliary vector. */
    for(i = 0;  i < AT_COUNT;  ++i)
	aux_info[i] = NULL;
    for(auxp = aux;  auxp->a_type != AT_NULL;  ++auxp) {
	if(auxp->a_type < AT_COUNT)
	    aux_info[auxp->a_type] = auxp;
    }

    /* Initialize and relocate ourselves. */
    assert(aux_info[AT_BASE] != NULL);
    init_rtld((caddr_t) aux_info[AT_BASE]->a_un.a_ptr);

    if(aux_info[AT_DEBUG] != NULL)	/* Set debugging level */
	debug = aux_info[AT_DEBUG]->a_un.a_val;

    __progname = "ld-elf.so.1";
    environ = env;

    trust = geteuid() == getuid() && getegid() == getgid();

    ld_bind_now = getenv("LD_BIND_NOW");
    if(trust) {
	ld_debug = getenv("LD_DEBUG");
	ld_library_path = getenv("LD_LIBRARY_PATH");
    }

    if(ld_debug != NULL && *ld_debug != '\0')
	debug = 1;
    dbg("%s is initialized", __progname);

    /*
     * Load the main program, or process its program header if it is
     * already loaded.
     */
    if(aux_info[AT_EXECFD] != NULL) {	/* Load the main program. */
	int fd = aux_info[AT_EXECFD]->a_un.a_val;
	dbg("loading main program");
	obj_main = map_object(fd);
	close(fd);
	if(obj_main == NULL)
	    return NULL;
    } else {				/* Main program already loaded. */
	const Elf32_Phdr *phdr;
	int phnum;
	caddr_t entry;

	dbg("processing main program's program header");
	assert(aux_info[AT_PHDR] != NULL);
	phdr = (const Elf32_Phdr *) aux_info[AT_PHDR]->a_un.a_ptr;
	assert(aux_info[AT_PHNUM] != NULL);
	phnum = aux_info[AT_PHNUM]->a_un.a_val;
	assert(aux_info[AT_PHENT] != NULL);
	assert(aux_info[AT_PHENT]->a_un.a_val == sizeof(Elf32_Phdr));
	assert(aux_info[AT_ENTRY] != NULL);
	entry = (caddr_t) aux_info[AT_ENTRY]->a_un.a_ptr;
	obj_main = digest_phdr(phdr, phnum, entry);
    }

    obj_main->path = xstrdup("main program");
    obj_main->mainprog = true;
    digest_dynamic(obj_main);

    /* Link the main program into the list of objects. */
    *obj_tail = obj_main;
    obj_tail = &obj_main->next;
    ++obj_main->refcount;

    dbg("loading needed objects");
    load_needed_objects(obj_main);

    dbg("relocating objects");
    relocate_objects(obj_main);

    dbg("doing copy relocations");
    do_copy_relocations(obj_main);

    if(aux_info[AT_BRK] != NULL) {
	/*
	 * We were invoked by the application-level testbed program.
	 * Since it is an a.out program, the address space is not laid
	 * out the way ELF expects it to be.  We have to manually poke
	 * the correct values into the variables that sbrk() uses for
	 * keeping track of memory allocation.
	 */
	Elf32_Addr brk = (Elf32_Addr) aux_info[AT_BRK]->a_un.a_ptr;

	patch_addr(".curbrk", brk);
	patch_addr(".minbrk", brk);
    }

    dbg("transferring control to program entry point = %p", obj_main->entry);

    /* Return with the entry point in %eax and the exit procedure in %edx. */

    return (long long)(Elf32_Addr) rtld_exit << 32 |
	(long long)(Elf32_Addr) obj_main->entry;
}

caddr_t
_rtld_bind(const Obj_Entry *obj, Elf32_Word reloff)
{
    const Elf32_Rel *rel;
    const Elf32_Sym *def;
    const Obj_Entry *defobj;
    Elf32_Addr *where;
    caddr_t target;

    rel = (const Elf32_Rel *) ((caddr_t) obj->pltrel + reloff);
    assert(ELF32_R_TYPE(rel->r_info) == R_386_JMP_SLOT);

    where = (Elf32_Addr *) (obj->relocbase + rel->r_offset);
    def = find_symdef(ELF32_R_SYM(rel->r_info), obj, &defobj, true);
    if(def == NULL)
	_exit(1);

    target = (caddr_t) (defobj->relocbase + def->st_value);

    dbg("\"%s\" in \"%s\" ==> %p in \"%s\"",
	defobj->strtab + def->st_name, basename(obj->path),
	target, basename(defobj->path));

    *where = (Elf32_Addr) target;
    return target;
}

static const char *
basename(const char *name)
{
    const char *p = strrchr(name, '/');
    return p != NULL ? p + 1 : name;
}

/*
 * Process a shared object's DYNAMIC section, and save the important
 * information in its Obj_Entry structure.
 */
static void
digest_dynamic(Obj_Entry *obj)
{
    const Elf32_Dyn *dynp;
    Needed_Entry **needed_tail = &obj->needed;
    const Elf32_Dyn *dyn_rpath = NULL;

    for(dynp = obj->dynamic;  dynp->d_tag != DT_NULL;  ++dynp) {
	switch(dynp->d_tag) {

	case DT_REL:
	    obj->rel = (const Elf32_Rel *) (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_RELSZ:
	    obj->relsize = dynp->d_un.d_val;
	    break;

	case DT_RELENT:
	    assert(dynp->d_un.d_val == sizeof(Elf32_Rel));
	    break;

	case DT_JMPREL:
	    obj->pltrel = (const Elf32_Rel *)
		(obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_PLTRELSZ:
	    obj->pltrelsize = dynp->d_un.d_val;
	    break;

	case DT_RELA:
	case DT_RELASZ:
	case DT_RELAENT:
	    assert(0);	/* Should never appear for i386 */
	    break;

	case DT_PLTREL:
	    assert(dynp->d_un.d_val == DT_REL);		/* For the i386 */
	    break;

	case DT_SYMTAB:
	    obj->symtab = (const Elf32_Sym *)
		(obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_SYMENT:
	    assert(dynp->d_un.d_val == sizeof(Elf32_Sym));
	    break;

	case DT_STRTAB:
	    obj->strtab = (const char *) (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_STRSZ:
	    obj->strsize = dynp->d_un.d_val;
	    break;

	case DT_HASH:
	    {
		const Elf32_Word *hashtab = (const Elf32_Word *)
		    (obj->relocbase + dynp->d_un.d_ptr);
		obj->nbuckets = hashtab[0];
		obj->nchains = hashtab[1];
		obj->buckets = hashtab + 2;
		obj->chains = obj->buckets + obj->nbuckets;
	    }
	    break;

	case DT_NEEDED:
	    assert(!obj->rtld);
	    {
		Needed_Entry *nep = NEW(Needed_Entry);
		nep->name = dynp->d_un.d_val;
		nep->next = NULL;

		*needed_tail = nep;
		needed_tail = &nep->next;
	    }
	    break;

	case DT_PLTGOT:
	    obj->got = (Elf32_Addr *) (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_TEXTREL:
	    obj->textrel = true;
	    break;

	case DT_SYMBOLIC:
	    obj->symbolic = true;
	    break;

	case DT_RPATH:
	    /*
	     * We have to wait until later to process this, because we
	     * might not have gotten the address of the string table yet.
	     */
	    dyn_rpath = dynp;
	    break;

	case DT_SONAME:
	    /* Not used by the dynamic linker. */
	    break;

	case DT_INIT:
	    obj->init = (void (*)(void)) (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_FINI:
	    obj->fini = (void (*)(void)) (obj->relocbase + dynp->d_un.d_ptr);
	    break;

	case DT_DEBUG:
	    /* FIXME - not implemented yet */
	    break;
	}
    }

    if(dyn_rpath != NULL)
	obj->rpath = obj->strtab + dyn_rpath->d_un.d_val;
}

/*
 * Process a shared object's program header.  This is used only for the
 * main program, when the kernel has already loaded the main program
 * into memory before calling the dynamic linker.  It creates and
 * returns an Obj_Entry structure.
 */
static Obj_Entry *
digest_phdr(const Elf32_Phdr *phdr, int phnum, caddr_t entry)
{
    Obj_Entry *obj = CNEW(Obj_Entry);
    const Elf32_Phdr *phlimit = phdr + phnum;
    const Elf32_Phdr *ph;
    int nsegs = 0;

    for(ph = phdr;  ph < phlimit;  ++ph) {
	switch(ph->p_type) {

	case PT_PHDR:
	    assert((const Elf32_Phdr *) ph->p_vaddr == phdr);
	    obj->phdr = (const Elf32_Phdr *) ph->p_vaddr;
	    obj->phsize = ph->p_memsz;
	    break;

	case PT_LOAD:
	    assert(nsegs < 2);
	    if(nsegs == 0) {	/* First load segment */
		obj->vaddrbase = round_down(ph->p_vaddr);
		obj->mapbase = (caddr_t) obj->vaddrbase;
		obj->relocbase = obj->mapbase - obj->vaddrbase;
		obj->textsize = round_up(ph->p_vaddr + ph->p_memsz) -
		    obj->vaddrbase;
	    } else {		/* Last load segment */
		obj->mapsize = round_up(ph->p_vaddr + ph->p_memsz) -
		    obj->vaddrbase;
	    }
	    ++nsegs;
	    break;

	case PT_DYNAMIC:
	    obj->dynamic = (const Elf32_Dyn *) ph->p_vaddr;
	    break;
	}
    }
    assert(nsegs == 2);

    obj->entry = entry;
    return obj;
}

/*
 * Process the special R_386_COPY relocations in the main program.  These
 * copy data from a shared object into a region in the main program's BSS
 * segment.
 *
 * Returns 0 on success, -1 on failure.
 */
static int
do_copy_relocations(Obj_Entry *dstobj)
{
    const Elf32_Rel *rellim = (const Elf32_Rel *)
	((caddr_t) dstobj->rel + dstobj->relsize);
    const Elf32_Rel *rel;

    assert(dstobj->mainprog);	/* COPY relocations are invalid elsewhere */

    for(rel = dstobj->rel;  rel < rellim;  ++rel) {
	if(ELF32_R_TYPE(rel->r_info) == R_386_COPY) {
	    void *dstaddr = (void *) (dstobj->relocbase + rel->r_offset);
	    const Elf32_Sym *dstsym =
		dstobj->symtab + ELF32_R_SYM(rel->r_info);
	    const char *name = dstobj->strtab + dstsym->st_name;
	    unsigned long hash = elf_hash(name);
	    size_t size = dstsym->st_size;
	    const void *srcaddr;
	    const Elf32_Sym *srcsym;
	    Obj_Entry *srcobj;

	    for(srcobj = dstobj->next;  srcobj != NULL;  srcobj = srcobj->next)
		if((srcsym = symlook_obj(name, hash, srcobj, false)) != NULL)
		    break;

	    if(srcobj == NULL) {
		error("Undefined symbol \"%s\" referenced from COPY"
		    " relocation in %s", name, dstobj->path);
		return -1;
	    }

	    srcaddr = (const void *) (srcobj->relocbase + srcsym->st_value);

	    memcpy(dstaddr, srcaddr, size);
	}
    }

    return 0;
}

/*
 * Hash function for symbol table lookup.  Don't even think about changing
 * this.  It is specified by the System V ABI.
 */
static unsigned long
elf_hash(const char *name)
{
    const unsigned char *p = (const unsigned char *) name;
    unsigned long h = 0;
    unsigned long g;

    while(*p != '\0') {
	h = (h << 4) + *p++;
	if((g = h & 0xf0000000) != 0)
	    h ^= g >> 24;
	h &= ~g;
    }
    return h;
}

/*
 * Error reporting function.  Use it like printf.
 *
 * This will eventually be changed so that it saves its error message in
 * a buffer, for subsequent retrieval via dlerror().  But for now, it
 * just prints a message and terminates the program.
 */
void
error(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    verrx(1, fmt, ap);
    va_end(ap);
}

/*
 * Find the library with the given name, and return its full pathname.
 * The returned string is dynamically allocated.  Generates an error
 * message and returns NULL if the library cannot be found.
 */
static char *
find_library(const char *name, const Obj_Entry *refobj)
{
    char *pathname;

    if(strchr(name, '/') != NULL) {	/* Hard coded pathname */
	if(name[0] != '/' && !trust) {
	    error("Absolute pathname required for shared object \"%s\"", name);
	    return NULL;
	}
	if(strncmp(name, SVR4_LIBDIR, SVR4_LIBDIRLEN) == 0 &&
	name[SVR4_LIBDIRLEN] == '/') {	/* In "/usr/lib" */
	    /* Map hard-coded "/usr/lib" onto our ELF library directory. */
	    pathname = xmalloc(strlen(name) + LIBDIRLEN - SVR4_LIBDIRLEN + 1);
	    strcpy(pathname, LIBDIR);
	    strcpy(pathname + LIBDIRLEN, name + SVR4_LIBDIRLEN);
	} else
	    pathname = xstrdup(name);

	return pathname;
    }

    dbg(" Searching for \"%s\"", name);

    if((pathname = search_library_path(name, refobj->rpath)) != NULL ||
       (pathname = search_library_path(name, ld_library_path)) != NULL ||
       (pathname = search_library_path(name, SVR4_LIBDIR)) != NULL)
	return pathname;

    error("Shared object \"%s\" not found", name);
    return NULL;
}

/*
 * Given a symbol number in a referencing object, find the corresponding
 * definition of the symbol.  Returns a pointer to the symbol, or NULL if
 * no definition was found.  Returns a pointer to the Obj_Entry of the
 * defining object via the reference parameter DEFOBJ_OUT.
 */
static const Elf32_Sym *
find_symdef(unsigned long symnum, const Obj_Entry *refobj,
    const Obj_Entry **defobj_out, bool in_plt)
{
    const Elf32_Sym *ref = refobj->symtab + symnum;
    const char *name = refobj->strtab + ref->st_name;
    const Elf32_Sym *def = NULL;

    if(refobj->symbolic) {	/* Look first in the referencing object */
	def = symlook_obj(name, elf_hash(name), refobj, in_plt);
	if(def != NULL)
	    *defobj_out = refobj;
    }
    if(def == NULL)		/* Look in all loaded objects */
	def = symlook(name, defobj_out, in_plt);

    if(def == NULL) {
	error("Undefined symbol \"%s\" referenced from %s",
	    name, refobj->path);
    }
    return def;
}

/*
 * Initialize the dynamic linker.  The argument is the address at which
 * the dynamic linker has been mapped into memory.  The primary task of
 * this function is to relocate the dynamic linker.
 */
static void
init_rtld(caddr_t mapbase)
{
    Obj_Entry obj_rtld;

    /* Conjure up an Obj_Entry structure for the dynamic linker. */

    memset(&obj_rtld, 0, sizeof obj_rtld);
    obj_rtld.rtld = true;
    obj_rtld.mapbase = mapbase;
    obj_rtld.relocbase = mapbase;
    get_got_address(obj_rtld.got);
    obj_rtld.dynamic = (const Elf32_Dyn *) (obj_rtld.mapbase + obj_rtld.got[0]);

    digest_dynamic(&obj_rtld);
    assert(obj_rtld.needed == NULL);
    assert(!obj_rtld.textrel);

    /*
     * Set up the obj_list pointer, so that rtld symbols can be found. 
     * This works even before relocating ourselves, because obj_list is
     * static.  (Its address is just an offset from the GOT pointer.)
     */

    obj_list = &obj_rtld;

    relocate_objects(&obj_rtld);

    obj_list = NULL;
    obj_tail = &obj_list;
}

/*
 * Given a shared object, traverse its list of needed objects, and load
 * each of them.  Returns 0 on success.  Generates an error message and
 * returns -1 on failure.
 */
static int
load_needed_objects(Obj_Entry *first)
{
    Obj_Entry *obj;

    for(obj = first;  obj != NULL;  obj = obj->next) {
	const Needed_Entry *needed;

	for(needed = obj->needed;  needed != NULL;  needed = needed->next) {
	    const char *name = obj->strtab + needed->name;
	    char *path = find_library(name, obj);
	    Obj_Entry *needed_obj;

	    if(path == NULL)
		return -1;

	    needed_obj = load_object(path);
	    if(needed_obj == NULL)
		return -1;		/* FIXME - cleanup */
	}
    }

    return 0;
}

/*
 * Load a shared object into memory, if it is not already loaded.
 * Returns a pointer to the Obj_Entry for the object.  Returns NULL
 * on failure.
 */
static Obj_Entry *
load_object(char *path)
{
    Obj_Entry *obj;

    for(obj = obj_list->next;  obj != NULL;  obj = obj->next)
	if(strcmp(obj->path, path) == 0)
	    break;

    if(obj == NULL) {	/* First use of this object, so we must map it in */
	int fd;

	if((fd = open(path, O_RDONLY)) == -1) {
	    error("Cannot open \"%s\"", path);
	    return NULL;
	}
	obj = map_object(fd);
	close(fd);
	if(obj == NULL) {
	    free(path);
	    return NULL;
	}

	obj->path = path;
	digest_dynamic(obj);

	*obj_tail = obj;
	obj_tail = &obj->next;

	dbg("  %p .. %p: %s", obj->mapbase,
	    obj->mapbase + obj->mapsize - 1, obj->path);
    } else
	free(path);

    ++obj->refcount;
    return obj;
}

/*
 * Patch a specified value into the 32-bit word designated by
 * the given symbol name.  Returns 0 on success, or -1 on failure.
 */
static int
patch_addr(const char *symbol, Elf32_Addr value)
{
    const Elf32_Sym *def;
    const Obj_Entry *defobj;
    Elf32_Addr *where;

    def = symlook(symbol, &defobj, false);
    if(def == NULL)
	return -1;

    where = (Elf32_Addr *) (defobj->relocbase + def->st_value);
    *where = value;

    dbg("patched \"%s\" to 0x%lx at %p", symbol, (long) value, where);
    return 0;
}

/*
 * Relocate newly-loaded shared objects.  The argument is a pointer to
 * the Obj_Entry for the first such object.  All objects from the first
 * to the end of the list of objects are relocated.  Returns 0 on success,
 * or -1 on failure.
 */
static int
relocate_objects(Obj_Entry *first)
{
    Obj_Entry *obj;

    for(obj = first;  obj != NULL;  obj = obj->next) {
	const Elf32_Rel *rellim;
	const Elf32_Rel *rel;

	if(obj->textrel) {
	    /* There are relocations to the write-protected text segment. */
	    if(mprotect(obj->mapbase, obj->textsize,
		PROT_READ|PROT_WRITE|PROT_EXEC) == -1) {
		error("Cannot unprotect text segment: %s", strerror(errno));
		return -1;
	    }
	}

	/* Process the non-PLT relocations. */
	rellim = (const Elf32_Rel *) ((caddr_t) obj->rel + obj->relsize);
	for(rel = obj->rel;  rel < rellim;  ++rel) {
	    Elf32_Addr *where = (Elf32_Addr *) (obj->relocbase + rel->r_offset);

	    switch(ELF32_R_TYPE(rel->r_info)) {

	    case R_386_NONE:
		break;

	    case R_386_32:
		{
		    const Elf32_Sym *def;
		    const Obj_Entry *defobj;

		    def = find_symdef(ELF32_R_SYM(rel->r_info), obj, &defobj,
			false);
		    if(def == NULL)
			return -1;

		    *where += (Elf32_Addr) (defobj->relocbase + def->st_value);
		}
		break;

	    case R_386_COPY:
		/*
		 * These are deferred until all other relocations have
		 * been done.  All we do here is make sure that the COPY
		 * relocation is not in a shared library.  They are allowed
		 * only in executable files.
		 */
		if(!obj->mainprog) {
		    error("Unexpected R_386_COPY relocation in shared library");
		    return -1;
		}
		break;

	    case R_386_GLOB_DAT:
		{
		    const Elf32_Sym *def;
		    const Obj_Entry *defobj;

		    def = find_symdef(ELF32_R_SYM(rel->r_info), obj, &defobj,
			false);
		    if(def == NULL)
			return -1;

		    *where = (Elf32_Addr) (defobj->relocbase + def->st_value);
		}
		break;

	    case R_386_RELATIVE:
		*where += (Elf32_Addr) obj->relocbase;
		break;

	    default:
		error("Unsupported relocation type %d in non-PLT relocations\n",
		    ELF32_R_TYPE(rel->r_info));
		return -1;
	    }
	}

	if(obj->textrel) {	/* Re-protected the text segment. */
	    if(mprotect(obj->mapbase, obj->textsize,
		PROT_READ|PROT_EXEC) == -1) {
		error("Cannot reprotect text segment: %s", strerror(errno));
		return -1;
	    }
	}

	/* Process the PLT relocations. */
	rellim = (const Elf32_Rel *) ((caddr_t) obj->pltrel + obj->pltrelsize);
	if(ld_bind_now != NULL && *ld_bind_now != '\0') {
	    /* Fully resolve procedure addresses now */
	    for(rel = obj->pltrel;  rel < rellim;  ++rel) {
		Elf32_Addr *where = (Elf32_Addr *)
		    (obj->relocbase + rel->r_offset);
		const Elf32_Sym *def;
		const Obj_Entry *defobj;

		assert(ELF32_R_TYPE(rel->r_info) == R_386_JMP_SLOT);

		def = find_symdef(ELF32_R_SYM(rel->r_info), obj, &defobj, true);
		if(def == NULL)
		    return -1;

		*where = (Elf32_Addr) (defobj->relocbase + def->st_value);
	    }
	} else {	/* Just relocate the GOT slots pointing into the PLT */
	    for(rel = obj->pltrel;  rel < rellim;  ++rel) {
		Elf32_Addr *where = (Elf32_Addr *)
		    (obj->relocbase + rel->r_offset);
		*where += (Elf32_Addr) obj->relocbase;
	    }
	}

	/* Set the special GOT entries. */
	obj->got[1] = (Elf32_Addr) obj;
	obj->got[2] = (Elf32_Addr) &_rtld_bind_start;
    }

    return 0;
}

/*
 * Cleanup procedure.  It will be called (by the atexit() mechanism) just
 * before the process exits.
 */
static void
rtld_exit(void)
{
    /* No cleanup needed yet */
}

static char *
search_library_path(const char *name, const char *path)
{
    size_t namelen = strlen(name);
    const char *p = path;

    if(p == NULL)
	return NULL;

    p += strspn(p, ":;");
    while(*p != '\0') {
	size_t len = strcspn(p, ":;");

	if(*p == '/' || trust) {
	    char *pathname;
	    const char *dir = p;
	    size_t dirlen = len;

	    if(dirlen == SVR4_LIBDIRLEN &&
	    strncmp(dir, SVR4_LIBDIR, SVR4_LIBDIRLEN) == 0) {
		/* Substitute our special ELF directory for "/usr/lib" */
		dir = LIBDIR;
		dirlen = LIBDIRLEN;
	    }

	    pathname = xmalloc(dirlen + 1 + namelen + 1);
	    strncpy(pathname, dir, dirlen);
	    pathname[dirlen] = '/';
	    strcpy(pathname + dirlen + 1, name);

	    dbg("  Trying \"%s\"", pathname);
	    if(access(pathname, F_OK) == 0)		/* We found it */
		return pathname;

	    free(pathname);
	}
	p += len;
	p += strspn(p, ":;");
    }

    return NULL;
}

/*
 * Find the definition of the given symbol name.  Returns a pointer to
 * the symbol, or NULL if no definition was found.  The defining object
 * is returned via the reference parameter OBJ_OUT.
 */
static const Elf32_Sym *
symlook(const char *name, const Obj_Entry **obj_out, bool in_plt)
{
    unsigned long hash = elf_hash(name);
    const Obj_Entry *obj;

    for(obj = obj_list;  obj != NULL;  obj = obj->next) {
	const Elf32_Sym *symp = symlook_obj(name, hash, obj, in_plt);
	if(symp != NULL) {
	    *obj_out = obj;
	    return symp;
	}
    }

    return NULL;
}

/*
 * Search the symbol table of a single shared object for a symbol of
 * the given name.  Returns a pointer to the symbol, or NULL if no
 * definition was found.
 *
 * The symbol's hash value is passed in for efficiency reasons; that
 * eliminates many recomputations of the hash value.
 */
static const Elf32_Sym *
symlook_obj(const char *name, unsigned long hash, const Obj_Entry *obj,
    bool in_plt)
{
    unsigned long symnum = obj->buckets[hash % obj->nbuckets];

    while(symnum != STN_UNDEF) {
	const Elf32_Sym *symp;
	const char *strp;

	assert(symnum < obj->nchains);
	symp = obj->symtab + symnum;
	assert(symp->st_name != 0);
	strp = obj->strtab + symp->st_name;

	if(strcmp(name, strp) == 0) {
	    if(symp->st_shndx != SHN_UNDEF ||
	    (!in_plt && symp->st_value != 0 &&
	    ELF32_ST_TYPE(symp->st_info) == STT_FUNC)) {
		return symp;
	    }
	}

	symnum = obj->chains[symnum];
    }

    return NULL;
}

/*
 * Non-mallocing printf, for use by malloc itself.
 * FIXME - This doesn't belong in this module.
 */
void
xprintf(const char *fmt, ...)
{
    char buf[256];
    va_list	ap;
    va_start(ap, fmt);

    vsprintf(buf, fmt, ap);
    (void)write(1, buf, strlen(buf));
    va_end(ap);
}
