/*
 * Copyright 1996 John D. Polstra.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by John Polstra.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *      $Id: run.c,v 1.2 1996/01/19 18:37:15 jdp Exp $
 */

/*
 * Testbed for loading and executing an ELF program.
 */

#include <assert.h>
#include <err.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "debug.h"
#include "rtld.h"

#define AUX_PTR(auxp, type, ptr) \
    ((auxp)->a_type = (type), (auxp)->a_un.a_ptr = (void *)(ptr), ++(auxp))
#define AUX_VAL(auxp, type, val) \
    ((auxp)->a_type = (type), (auxp)->a_un.a_val = (val), ++(auxp))

static Obj_Entry *map_file(const char *);
static void usage(const char *);

int
main(int argc, char *argv[], char *env[])
{
    int opt;
    char *filename;
    Obj_Entry *obj_prog;
    Obj_Entry *obj_rtld;
    size_t framesize;
    Elf32_Word *frame;
    char **src;
    char **dst;
    auxv_t *auxp;
    auxv_t *aux_brk;
    caddr_t brk_start;

    while((opt = getopt(argc, argv, "d")) != EOF) {
	switch(opt) {

	case 'd':	/* Enable debugging messages */
	    debug = 1;
	    break;

	default:
	    usage(argv[0]);
	    break;
	}
    }

    if(optind >= argc)
	usage(argv[0]);

    argv += optind;
    argc -= optind;
    filename = argv[0];

    dbg("Mapping main program");
    obj_prog = map_file(filename);
    dbg("Mapping rtld");
    obj_rtld = map_file(LDSO);

    /*
     * Build up a fake stack frame, as it would appear at the start of
     * the program.  We could write a bunch of assembly language code to
     * do this, but instead we cheat and use alloca() to grab some space
     * on the runtime stack.
     */
    framesize = 0;
    framesize += sizeof(Elf32_Word);		/* argc */
    framesize += (argc+1) * sizeof(char *);	/* Arguments and NULL */
    src = env;
    while(*src++ != NULL)			/* Count environment */
	;
    framesize += (src-env) * sizeof(char *);	/* Environment and NULL */
    framesize += AT_COUNT * sizeof(auxv_t);	/* More than enough */

    frame = (Elf32_Word *) alloca(framesize);

    *frame = argc;
    dst = (char **) (frame + 1);
    src = argv;
    while((*dst++ = *src++) != NULL)	/* Copy arguments */
	;
    src = env;
    while((*dst++ = *src++) != NULL)	/* Copy environment */
	;
    auxp = (auxv_t *) dst;
    AUX_PTR(auxp, AT_PHDR, obj_prog->phdr);
    AUX_VAL(auxp, AT_PHENT, sizeof(Elf32_Phdr));
    AUX_VAL(auxp, AT_PHNUM, obj_prog->phsize / sizeof(Elf32_Phdr));
    AUX_PTR(auxp, AT_ENTRY, obj_prog->entry);
    AUX_PTR(auxp, AT_BASE, obj_rtld->mapbase);
    AUX_VAL(auxp, AT_FLAGS, 0);
    AUX_VAL(auxp, AT_PAGESZ, PAGESIZE);

    aux_brk = auxp;
    AUX_PTR(auxp, AT_BRK, NULL);	/* Will be filled in later */

    AUX_VAL(auxp, AT_DEBUG, debug);
    AUX_VAL(auxp, AT_NULL, 0);

    assert((char *) auxp - (char *) frame <= framesize);

    /* Jump to the rtld entry point, never to return. */

    dbg("Transferring control to rtld");
    fflush(stdout);
    aux_brk->a_un.a_ptr = sbrk(0);
    __asm__ volatile ("movl %0,%%esp; jmp *%1"
		      : /* No outputs */
		      : "rm"(frame), "rm"(obj_rtld->entry)
		      : "%esp");

    assert(0);	/* We should never get here */
}

void
error(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    verrx(1, fmt, ap);
    va_end(ap);
}

static Obj_Entry *
map_file(const char *name)
{
    int fd;
    Obj_Entry *obj;

    if((fd = open(name, O_RDONLY)) == -1)
	err(1, "Cannot open \"%s\"", name);
    if((obj = map_object(fd)) == NULL)
	errx(1, "map_object failed");
    obj->path = xstrdup(name);

    dbg("  %p .. %p: %s", obj->mapbase,
	obj->mapbase + obj->mapsize - 1, obj->path);
    dbg("  entry point = %p", obj->entry);

    close(fd);
    return obj;
}

static void
usage(const char *progname)
{
    const char *basename;

    if((basename = strrchr(progname, '/')) != NULL)
	++basename;
    else
	basename = progname;

    errx(1, "Usage: %s [-d] elf-program [arguments ...]", basename);
}
